'''Publish to a channel

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import datetime
import functools
import json
import os
import random
import tarfile
import uuid

import click
import uvloop

from cobras.client.client import client
from cobras.client.credentials import (createCredentials, getDefaultRoleForApp,
                                      getDefaultSecretForApp)
from cobras.client.publish import computeEventTimeDeltas
from cobras.common.apps_config import PUBSUB_APPKEY
from cobras.runner.superuser import preventRootUsage

root = os.path.dirname(os.path.realpath(__file__))
dataDir = os.path.join(root, '..', '..', 'data')
DEFAULT_PATH = os.path.join(dataDir, 'publish.jsonl')
DEFAULT_BATCH_PATH = os.path.join(dataDir, 'niso_events.tar.bz')


# Default channel when pushing from redis
DEFAULT_CHANNEL = 'sms_republished_v1_neo'

DEFAULT_URL = f'ws://127.0.0.1:8765/v2?appkey={PUBSUB_APPKEY}'


async def sendEvent(websocket, channel, event, connectionId, verbose):
    data = json.loads(event)
    action = data.get('action')

    if action is None or action != 'rtm/publish':
        publishPdu = {
            "action": "rtm/publish",
            "body": {
                "channel": channel,
                "message": data
            }
        }
        line = json.dumps(publishPdu)
    else:
        data['body']['channel'] = channel
        line = json.dumps(data)

    if verbose:
        now = datetime.datetime.now().strftime("%H:%M:%S.%f")
        print(f"[{now}][{connectionId}] > {line}")

    # await websocket.send(line.encode('utf8'))
    await websocket.send(line)


async def clientCallback(websocket, **args):
    item = args['item']
    channel = args['channel']
    verbose = args['verbose']
    delay = args['delay']
    repeat = args['repeat']

    if not item:  # FIXME: how can this happen ?
        return

    while True:
        connectionId = uuid.uuid4().hex[:8]
        eventsWithDeltas = computeEventTimeDeltas(item)

        N = len(eventsWithDeltas)
        now = datetime.datetime.now().strftime("%H:%M:%S.%f")
        print(f"[{now}][{connectionId}] {N} events")

        for event, deltaMs in eventsWithDeltas:
            await sendEvent(websocket, channel, event, connectionId, verbose)

            if deltaMs != 0:
                print(f'sleeping for {deltaMs} ms')
                await asyncio.sleep(deltaMs / 1000)
            elif delay:
                print(f'sleeping for {delay} ms')
                await asyncio.sleep(delay)

        if not repeat:
            break


async def publishTask(url, credentials, items, channel,
                      verbose, repeat, delay):
    tasks = []
    for item in items:
        publishClientCallback = functools.partial(clientCallback,
                                                  item=item,
                                                  channel=channel,
                                                  verbose=verbose,
                                                  repeat=repeat,
                                                  delay=delay)

        task = asyncio.ensure_future(
            client(url, credentials, publishClientCallback))
        tasks.append(task)

    done, pending = await asyncio.wait(
        tasks,
        return_when=asyncio.ALL_COMPLETED,
    )
    for task in pending:
        task.cancel()


def buildItemList(path, limit):
    if tarfile.is_tarfile(path):
        items = []
        with tarfile.open(path, 'r:*') as tar:
            print('Extracting tarball...')
            members = tar.getmembers()
            for i, member in enumerate(members):
                if member.isfile():
                    f = tar.extractfile(member.name)
                    items.append(f.read().decode('utf-8'))

                if i > limit:
                    break

        # Randomize those input files
        random.shuffle(items)
    else:
        with open(path) as f:
            item = f.read()
        items = [item]

    return items


def run(url, channel, path, credentials, verbose, repeat, delay, limit):
    items = buildItemList(path, limit)
    if len(items) == 0:
        print('Empty input file')
        return

    print(f'Processing {len(items)} items')

    asyncio.get_event_loop().run_until_complete(
        publishTask(url, credentials, items, channel, verbose, repeat, delay))


@click.command()
@click.option('--url', default=DEFAULT_URL)
@click.option('--channel', default=DEFAULT_CHANNEL)
@click.option('--path', default=DEFAULT_PATH)
@click.option('--role', default=getDefaultRoleForApp('pubsub'))
@click.option('--secret', default=getDefaultSecretForApp('pubsub'))
@click.option('--verbose', is_flag=True)
@click.option('--repeat', is_flag=True)
@click.option('--batch', is_flag=True)
@click.option('--batch_events_path', envvar='COBRA_PUBLISH_BATCH_EVENTS_PATH',
              help='An archive (tar.gz, etc...) of events files')
@click.option('--limit', default=256)
@click.option('--delay', default=0.1)
def publish(url, channel, path, role, secret, batch, batch_events_path,
            limit, verbose, repeat, delay):
    '''Publish to a channel
    '''

    preventRootUsage()
    uvloop.install()

    if batch:
        path = batch_events_path

    credentials = createCredentials(role, secret)

    run(url, channel, path, credentials, verbose, repeat, delay, limit)



'''Setup cobra

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''
import click

from cobras.common.apps_config import AppsConfig, getDefaultAppsConfigPath
from cobras.runner.superuser import preventRootUsage


@click.command()
def init():
    '''Setup cobra
    '''

    preventRootUsage()

    path = getDefaultAppsConfigPath()
    appsConfig = AppsConfig(path)
    appsConfig.generateDefaultConfig()



'''Redis subscriber

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import traceback
from abc import ABC, abstractmethod

import ujson

from cobras.server.redis_connections import RedisConnections


class RedisSubscriberMessageHandlerClass(ABC):
    def __init__(self, args):
        pass  # pragma: no cover

    @abstractmethod
    def log(self, msg):
        pass  # pragma: no cover

    @abstractmethod
    async def on_init(self):
        pass  # pragma: no cover

    @abstractmethod
    async def handleMsg(self, msg: dict, payloadSize: int) -> bool:
        return True  # pragma: no cover


async def redisSubscriber(redisConnections: RedisConnections,
                          pattern: str,
                          messageHandlerClass: RedisSubscriberMessageHandlerClass,  # noqa
                          obj):
    # Create connection
    connection = await redisConnections.create(pattern, useAioRedis=False)

    # Create subscriber.
    subscriber = await connection.start_subscribe()

    # Subscribe to channel.
    await subscriber.subscribe([pattern])

    messageHandler = messageHandlerClass(obj)
    await messageHandler.on_init(connection)

    try:
        # wait for incoming events.
        while True:
            reply = await subscriber.next_published()
            msg = reply.value

            payloadSize = len(msg)
            msg = ujson.loads(msg)
            ret = await messageHandler.handleMsg(msg, payloadSize)
            if not ret:
                break

    except asyncio.CancelledError:
        messageHandler.log('Cancelling redis subscription')
        raise

    except Exception as e:
        messageHandler.log(e)
        messageHandler.log(
            'Generic Exception caught in {}'.format(traceback.format_exc()))

    finally:
        messageHandler.log('Closing redis subscription')

        # When finished, close the connection.
        connection.close()


def runSubscriber(redisConnections: RedisConnections,
                  channel: str, messageHandlerClass, obj=None):
    asyncio.get_event_loop().run_until_complete(
        redisSubscriber(redisConnections, channel,
                        messageHandlerClass, obj))



'''Handle set of redis connections. Used for sharding connections.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import logging
from urllib.parse import urlparse

import asyncio_redis
from uhashring import HashRing

from cobras.server.redis_publisher import create_redis_publisher


class RedisConnections():
    def __init__(self, urls: str, password) -> None:
        self.urls = urls.split(';')
        self.password = password
        if password == '':
            self.password = None

        # create a consistent hash ring
        # https://www.paperplanes.de/2011/12/9/the-magic-of-consistent-hashing.html
        self.hr = HashRing(nodes=self.urls)

    async def create(self, appChannel=None, useAioRedis=True):
        url = self.hashChannel(appChannel)
        logging.debug(f'Hashing {appChannel} to url -> {url}')

        netloc = urlparse(url).netloc
        host, _, port = netloc.partition(':')
        if port:
            port = int(port)
        else:
            port = 6379

        if useAioRedis:
            redis = await create_redis_publisher(host, port, self.password)
        else:
            redis = \
                await asyncio_redis.Connection.create(host=host, port=port,
                                                      password=self.password)

        return redis

    def hashChannel(self, appChannel: str):
        return self.hr.get_node(appChannel)



'''Health check to validate that a server is ok.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import json
import uuid

from cobras.client.client import subscribeClient
from cobras.client.credentials import createCredentials
from cobras.common.apps_config import HEALTH_APPKEY


def getDefaultHealthCheckChannel():
    randomKey = uuid.uuid4().hex[:8]
    return f'sms_health_check_channel_{randomKey}'


def getDefaultHealthCheckHttpUrl(host=None, port=None):
    if host is None:
        host = '127.0.0.1'
    if port is None:
        port = 8765

    return f'http://{host}:{port}/health/'


def getDefaultHealthCheckUrl(host=None, port=None):
    if host is None:
        host = '127.0.0.1'
    if port is None:
        port = 8765

    return f'ws://{host}:{port}/v2?appkey={HEALTH_APPKEY}'


def healthCheck(url, role, secret, channel):
    '''Perform a health check'''

    class MessageHandlerClass:
        def __init__(self, websocket, args):
            self.websocket = websocket
            self.channel = args['channel']
            self.content = args['content']
            self.subscriptionId = self.channel
            self.msg = None

        async def on_init(self):
            await self.publishHealthMsg()

        async def unsubscribe(self):
            # Unsubscribe
            unsubscribePdu = {
                "action": "rtm/unsubscribe",
                "body": {
                    "subscription_id": self.subscriptionId
                }
            }

            data = json.dumps(unsubscribePdu)
            print(f"> {data}")
            await self.websocket.send(data)
            await self.websocket.recv()  # needed ?

        async def publishHealthMsg(self):
            publishPdu = {
                "action": "rtm/publish",
                "body": {
                    "channel": self.channel,
                    "message": {
                        "position": "1519190184:547873030411",  # FIXME
                        "messages": [self.content]
                    }
                }
            }

            data = json.dumps(publishPdu)
            print(f"> {data}")
            await self.websocket.send(data)

        async def handleMsg(self, msg: str) -> bool:
            print('Received', msg)
            self.parsedMessage = json.loads(msg)

            await self.unsubscribe()

            # The server should not send this message to us
            await self.publishHealthMsg()
            return False

    credentials = createCredentials(role, secret)

    magicNumber = 666
    refAndroidId = uuid.uuid4().hex
    content = {
        'device': {
            'game': 'test',
            'android_id': refAndroidId
        },
        'magic': magicNumber
    }

    fsqlFilter = f"""select magic,device.android_id
                     from `{channel}` where
                         device.game = 'test' AND
                         device.android_id = '{refAndroidId}' AND
                         magic = {magicNumber}
    """
    messageHandler = asyncio.get_event_loop().run_until_complete(
        subscribeClient(url, credentials, channel,
                        fsqlFilter, MessageHandlerClass,
                        {'channel': channel, 'content': content}))

    data = messageHandler.parsedMessage
    messages = data['body']['messages']
    message = messages[0]
    assert message['device.android_id'] == refAndroidId
    assert magicNumber == message['magic']



'''Compute our version number

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import os

from cobras.common.memoize import memoize


@memoize
def getVersion():
    '''Get our own version using pkg_resources'''

    root = os.path.dirname(os.path.realpath(__file__))
    dataDir = os.path.join(root, '..', '..', '..')
    path = os.path.join(dataDir, 'DOCKER_VERSION')
    with open(path) as f:
        return f.read().replace('\n', '')



'''Stream SQL processor, to filter and transform messages when subcribing.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import collections
import logging

StreamSQLExpression = collections.namedtuple(
    'StreamSQLExpression', [
        'components', 'val', 'length',
        'equalExpression',
        'likeExpression',
        'differentExpression',
        'largerThanExpression',
        'lowerThanExpression'
    ]
)


class InvalidStreamSQLError(Exception):
    pass


class StreamSqlFilter:
    def __init__(self, sql_filter):
        '''
        Support expression on strings, bool, int
        Support AND and OR expressions

        No parenthesis support
        '''

        self.emptyFilter = False
        self.channel = None

        if sql_filter is None:
            raise InvalidStreamSQLError()

        # Basic validation
        tokens = [token.strip() for token in sql_filter.split()]

        validSql = len(tokens) >= 4 and \
            tokens[0].lower() == 'select' and \
            tokens[2].lower() == 'from'

        if not validSql:
            raise InvalidStreamSQLError()

        # Parse glob
        if tokens[1] == '*':
            self.fields = None
        else:
            self.fields = tokens[1]

        # Parse channel name
        self.channel = tokens[3].replace('`', '')

        if len(tokens) == 4:
            self.emptyFilter = True
            return

        if 'where' in sql_filter:
            whereStatement = 'where'
        elif 'WHERE' in sql_filter:
            whereStatement = 'WHERE'
        else:
            self.emptyFilter = True
            return

        _, _, sql_filter = sql_filter.partition(whereStatement)

        sql_filter = sql_filter.replace('\n', ' ')

        # Since we don't have a legit lexer, we need spaces around our
        # OR and AND ...
        # FIXME: we should figure out how to use ply
        self.andExpr = True
        if ' AND ' in sql_filter or ' OR ' in sql_filter:
            if ' AND ' in sql_filter:
                exprs = sql_filter.split(' AND ')
            elif ' OR ' in sql_filter:
                exprs = sql_filter.split(' OR ')
                self.andExpr = False

            exprs = [expr.strip() for expr in exprs]
            self.expressions = [self.buildExpression(expr) for expr in exprs]
        else:
            self.expressions = [self.buildExpression(sql_filter)]

    def buildExpression(self, text):
        '''
        FIXME: support lower than and greater than for ints
               support float ?
        '''

        equalExpression = False
        likeExpression = False
        differentExpression = False
        largerThanExpression = False
        lowerThanExpression = False

        if ' LIKE ' in text:
            jsonExpr, _, val = text.partition(' LIKE ')
            likeExpression = True
        elif ' != ' in text:
            jsonExpr, _, val = text.partition(' != ')
            differentExpression = True
        elif ' = ' in text:
            jsonExpr, _, val = text.partition('=')
            equalExpression = True
        elif ' > ' in text:
            jsonExpr, _, val = text.partition('>')
            largerThanExpression = True
        elif ' < ' in text:
            jsonExpr, _, val = text.partition('<')
            lowerThanExpression = True
        else:
            raise InvalidStreamSQLError('Invalid operand')

        self.jsonExpr = jsonExpr.strip()

        val = val.strip()
        if val.startswith("'") and val.endswith("'"):
            val = val.replace("'", '')
        elif val == 'true' or val == 'false':
            val = val == 'true'
        else:
            try:
                val = int(val)
            except ValueError:
                raise InvalidStreamSQLError()

        components = self.jsonExpr.split('.')
        return StreamSQLExpression(components, val, len(components),
                                   equalExpression,
                                   likeExpression,
                                   differentExpression,
                                   largerThanExpression,
                                   lowerThanExpression)

    def matchExpression(self, msg, expression):
        if expression.length == 2:
            val = msg.get(expression.components[0], {}).get(expression.components[1])  # noqa
            return self.matchOperand(val, expression)
        elif expression.length == 1:
            val = msg.get(expression.components[0])
            return self.matchOperand(val, expression)
        else:
            msgCopy = dict(msg)

            for component in expression.components:
                msgCopy = msgCopy.get(component, {})

            return self.matchOperand(msgCopy, expression)

    def matchOperand(self, val, expression):
        if expression.equalExpression:
            return expression.val == val
        elif expression.likeExpression:
            return expression.val in val
        elif expression.differentExpression:
            return expression.val != val
        elif expression.largerThanExpression:
            return val > expression.val
        elif expression.lowerThanExpression:
            return val < expression.val

    def match(self, msg):
        if self.emptyFilter:
            return self.transform(msg)

        if isinstance(msg, list):
            msg = msg[0]

        if not isinstance(msg, dict):
            logging.error('Bad type for {}'.format(msg))
            return False

        if self.andExpr:
            if not all(self.matchExpression(msg, expression) for expression in self.expressions):  # noqa
                return False
            else:
                return self.transform(msg)
        else:
            if not any(self.matchExpression(msg, expression) for expression in self.expressions):  # noqa
                return False
            else:
                return self.transform(msg)

    def transform(self, msg):
        '''extract a subfield'''

        if self.fields is None:
            return msg

        ret = {}
        fields = self.fields.split(',')
        for field in fields:
            subtree = dict(msg)
            fieldsComponents = field.split('.')
            for component in fieldsComponents:
                subtree = subtree.get(component, {})

            ret.update({field: subtree})

        return ret


def match_stream_sql_filter(sql_filter, msg):
    try:
        f = StreamSqlFilter(sql_filter)
        return f.match(msg)
    except InvalidStreamSQLError:
        return False



'''Tools used when running as superuser (which we should almost never do)

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import os
import sys

import click


def preventRootUsage():
    ''' Make sure a root user cannot run this command'''

    if os.geteuid() == 0:
        msg = 'Cobra cannot be run as root / sudo. Exiting'
        click.secho(msg, fg='red')
        sys.exit(1)



'''Cobra main driver.
   Calls into sub commands like git.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import click
import coloredlogs

from cobras.common.version import getVersion
from cobras.runner.commands.admin import admin
from cobras.runner.commands.health import health
from cobras.runner.commands.init import init
from cobras.runner.commands.monitor import monitor
from cobras.runner.commands.publish import publish
from cobras.runner.commands.redis_subscribe import redis_subscribe
from cobras.runner.commands.run import run
from cobras.runner.commands.secret import secret
from cobras.runner.commands.subscribe import subscribe

coloredlogs.install(level='WARNING')


@click.group()
@click.version_option()
def cli():
    '''Cobra is a realtime messaging server using
    Python3, WebSockets and Redis.'''
    pass


cli.add_command(run)
cli.add_command(publish)
cli.add_command(subscribe)
cli.add_command(redis_subscribe)
cli.add_command(monitor)
cli.add_command(health)
cli.add_command(secret)
cli.add_command(admin)
cli.add_command(init)



'''Monitor cobra

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import click
import uvloop

from cobras.client.credentials import (createCredentials, getDefaultRoleForApp,
                                      getDefaultSecretForApp)
from cobras.client.monitor import runMonitor
from cobras.common.apps_config import STATS_APPKEY
from cobras.runner.superuser import preventRootUsage

DEFAULT_URL = f'ws://127.0.0.1:8765/v2?appkey={STATS_APPKEY}'


@click.command()
@click.option('--url', default=DEFAULT_URL)
@click.option('--role', default=getDefaultRoleForApp('stats'))
@click.option('--secret', default=getDefaultSecretForApp('stats'))
@click.option('--raw', is_flag=True)
@click.option('--tidy', is_flag=True)
@click.option('--subscribers', is_flag=True)
@click.option('--role_filter')
def monitor(url, role, secret, raw, role_filter, tidy, subscribers):
    '''Monitor cobra
    '''

    preventRootUsage()
    uvloop.install()

    credentials = createCredentials(role, secret)
    displayNodes = not tidy
    runMonitor(url, credentials, raw, role_filter, displayNodes, subscribers)



'''Custom 'publish only' redis client which uses redis PIPELINING

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio


class RedisPublisher(object):
    '''
    See https://redis.io/topics/mass-insert
    '''
    def __init__(self, host, port, password, verbose=False):
        self.host = host
        self.port = port
        self.password = password
        self.verbose = verbose
        self.reset()
        self.lock = asyncio.Lock()

    def reset(self):
        self.publishCount = 0

    async def connect(self):
        async with self.lock:
            print(f'Opening connection to redis at {self.host}:{self.port}')
            self.reader, self.writer = await asyncio.open_connection(
                self.host, self.port)

        if self.password:
            self.writer.write(b'*2\r\n')
            self.writeString(b'AUTH')

            password = self.password
            if not isinstance(password, bytes):
                password = password.encode('utf8')

            self.writeString(password)
            await self.execute()

    async def execute(self):
        async with self.lock:
            await self.writer.drain()

            results = []

            # read until we get something
            for i in range(self.publishCount):
                line = await self.reader.readline()
                results.append(line)

                # FIXME: proper error handling !!!
                if self.verbose:
                    print(f'Received: {line.decode()!r}')

                if 'NOAUTH Authentication required' in line.decode():
                    raise ValueError('Authentication failed')

            self.reset()

            return results

    def close(self):
        print('Close the connection')
        self.reset()
        self.writer.close()

    async def wait_closed(self):
        await self.writer.wait_closed()

    def publish(self, channel, msg):
        self.publishCount += 1

        if not isinstance(channel, bytes):
            channel = channel.encode('utf8')

        if not isinstance(msg, bytes):
            msg = msg.encode('utf8')

        self.writer.write(b'*3\r\n')
        self.writeString(b'PUBLISH')
        self.writeString(channel)
        self.writeString(msg)

    def writeString(self, data: bytes):
        # import pdb; pdb.set_trace()
        # data = data.encode('utf-8')

        self.writer.write(b'$%d\r\n' % len(data))
        self.writer.write(data)
        self.writer.write(b'\r\n')


async def create_redis_publisher(host, port, password, verbose=False):
    client = RedisPublisher(host, port, password, verbose)
    await client.connect()
    return client



'''Subscribe to a channel

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import json

import click
import uvloop

from cobras.client.client import subscribeClient
from cobras.client.credentials import (createCredentials, getDefaultRoleForApp,
                                      getDefaultSecretForApp)
from cobras.common.apps_config import PUBSUB_APPKEY
from cobras.common.throttle import Throttle
from cobras.runner.superuser import preventRootUsage

DEFAULT_URL = f'ws://127.0.0.1:8765/v2?appkey={PUBSUB_APPKEY}'


class MessageHandlerClass:
    def __init__(self, websockets, args):
        self.cnt = 0
        self.cntPerSec = 0
        self.throttle = Throttle(seconds=1)
        self.verbose = args['verbose']

    async def on_init(self):
        pass

    async def handleMsg(self, msg: str) -> bool:
        self.cnt += 1
        self.cntPerSec += 1

        if self.verbose >= 1:
            data = json.loads(msg)
            print(data['body']['messages'][0])

        if self.throttle.exceedRate():
            return True

        print(f"#messages {self.cnt} msg/s {self.cntPerSec}")
        self.cntPerSec = 0

        return True


@click.command()
@click.option('--url', default=DEFAULT_URL)
@click.option('--role', default=getDefaultRoleForApp('pubsub'))
@click.option('--secret', default=getDefaultSecretForApp('pubsub'))
@click.option('--channel', default='sms_republished_v1_neo')
@click.option('--verbose', '-v', count=True)
@click.option('--stream_sql')
def subscribe(url, role, secret, channel, stream_sql, verbose):
    '''Subscribe to a channel
    '''

    preventRootUsage()
    uvloop.install()

    credentials = createCredentials(role, secret)

    asyncio.get_event_loop().run_until_complete(
            subscribeClient(url, credentials, channel,
                            stream_sql, MessageHandlerClass,
                            {'verbose': verbose}))



'''App configuration, persisted in a yaml file. Credentials + some config

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import base64
import logging
import os
from pathlib import Path
from random import getrandbits, randint

import yaml

STATS_APPKEY = '_stats'
HEALTH_APPKEY = '_health'
ADMIN_APPKEY = '_admin'
PUBSUB_APPKEY = '_pubsub'


class AppsConfig():
    def __init__(self, path: str) -> None:
        self.path = path
        self.data = {}
        self.apps = {}

        if os.path.exists(self.path):
            with open(self.path) as f:
                self.data = yaml.load(f.read(), Loader=yaml.FullLoader) or {}
                self.apps = self.data.get('apps')

            self.validateConfig()
        else:
            msg = f'Apps config file does not exists: "{path}". '
            msg += 'Use `cobra init` to create a default config file'
            logging.warning(msg)

    def validateConfig(self):
        if self.apps is None:
            raise ValueError(f'No apps present in config file')

        for app in self.apps:
            for roleName, role in self.apps[app]['roles'].items():
                if not isinstance(role, dict):
                    raise ValueError(f'role "{roleName}" is not a dict')

                if role.get('secret') is None:
                    raise ValueError(f'role "{roleName}" is missing a secret')

    def isAppKeyValid(self, appkey: str) -> bool:
        return self.apps.get(appkey) is not None

    def getRoleSecret(self, appkey: str, role: str) -> bytes:
        if not self.isAppKeyValid(appkey):
            logging.warning(f'Missing appkey: "{appkey}"')
            raise KeyError

        roles = self.apps.get(appkey, {}).get('roles')
        if roles is None:
            logging.warning(f'Missing roles')
            raise KeyError

        secret = roles.get(role, {}).get('secret')
        if secret is None:
            logging.warning(f'Missing role: "{role}"')
            raise KeyError

        return secret.encode('ascii')

    def isBatchPublishEnabled(self, appkey: str) -> bytes:
        if not self.isAppKeyValid(appkey):
            logging.warning(f'Missing appkey: "{appkey}"')
            raise KeyError

        batchPublish = self.apps.get(appkey, {}).get('batch_publish', False)
        return batchPublish

    def getBatchPublishSize(self):
        return self.data.get('batch_publish_size', -1)

    def generateDefaultConfig(self):
        self.data['apps'] = {}

        for app in [STATS_APPKEY, HEALTH_APPKEY, ADMIN_APPKEY, PUBSUB_APPKEY]:

            # Create an app with a user/role
            self.data['apps'][app] = {}
            self.data['apps'][app]['roles'] = {}

            role = app[1:]  # remove the leading _

            self.data['apps'][app]['roles'][role] = {
                'secret': genSecret(),
                'permissions': ['subscribe', 'publish', 'admin']
            }

        # write to disk
        with open(self.path, 'w') as f:
            yaml.dump(self.data, f, default_flow_style=False)

        # dump to the console (we could pretty-print)
        with open(self.path) as f:
            print(f.read())

    def getDefaultRoleForApp(self, app) -> str:
        try:
            roles = self.data['apps'][f'_{app}'].get('roles', {})
        except KeyError:
            return ''

        try:
            roleName = list(roles.keys())[0]
        except IndexError:
            return ''

        return roleName

    def getDefaultSecretForApp(self, app) -> str:
        try:
            roles = self.data['apps'][f'_{app}'].get('roles', {})
        except KeyError:
            return ''

        try:
            roleName = list(roles.keys())[0]
        except IndexError:
            return ''

        role = roles.get(roleName, {})
        return role.get('secret', '')


def getDefaultAppsConfigPath():
    path = Path.home() / '.cobra.yaml'
    return str(path)


def generateNonce():
    bits = getrandbits(64)
    return base64.b64encode(bytearray(str(bits), 'utf8')).decode('ascii')


def genSecret():
    lowerChars = 'abcdef'
    upperChars = 'ABCDEF'
    nums = '0123456789'

    chars = lowerChars + upperChars + nums
    N = len(chars)

    secret = ''
    for _ in range(32):
        c = randint(0, N-1)
        secret += chars[c]

    return secret



'''Publish jobs gets enqueued into a multi-publisher, which publish them
to redis in a batch fashion, using a pipeline.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.

https://redis.io/topics/pipelining
Weird leaks https://bugs.python.org/issue31620
'''

import asyncio


class PipelinedPublisher():
    def __init__(self, redis, batchSize=None):
        self.redis = redis
        self.queue = asyncio.Queue()
        self.batchSize = batchSize or 100
        self.lock = asyncio.Lock()

    async def publishAll(self):
        async with self.lock:
            while True:
                try:
                    job = self.queue.get_nowait()
                except asyncio.QueueEmpty:
                    break
                else:
                    self.publish(job)
                    self.queue.task_done()

        await self.redis.execute()

    def enqueue(self, job):
        self.queue.put_nowait(job)

    def publish(self, job):
        appkey, channel, data = job
        appChannel = '{}::{}'.format(appkey, channel)
        self.redis.publish(appChannel, data)

    async def flush(self):
        await self.redis.execute()

    async def publishNow(self, job):
        async with self.lock:
            self.publish(job)
            await self.flush()

    async def push(self, job, batchPublish=False):
        if not batchPublish:
            await self.publishNow(job)
            return

        self.enqueue(job)

        if self.queue.qsize() >= self.batchSize:
            await self.publishAll()



'''Track memory usage

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.

https://blog.fugue.co/2017-03-06-diagnosing-and-fixing-memory-leaks-in-python.html  # noqa
'''

import asyncio
import gc
import sys
import tracemalloc
from tracemalloc import Filter

import byteformat
import psutil

DEFAULT_DURATION = 5


class MemoryDebugger():
    def __init__(self, duration=None, steps=None, mode=None):
        tracemalloc.start(10)

        self.duration = duration
        if duration is None:
            self.duration = DEFAULT_DURATION

        self.mode = mode
        if mode is None:
            self.mode = 'filename'  # or traceback, or lineno
            self.mode = 'traceback'  # or traceback, or lineno

        self.steps = steps
        self.stop = False

        self.snapshots = []
        self.filters = [Filter(inclusive=True, filename_pattern="**")]
        self.p = psutil.Process()

        self.printAllTasks = False

    def collect_stats(self):
        self.snapshots.append(tracemalloc.take_snapshot())

        if len(self.snapshots) < 2:
            return
        else:
            # Only keep the last 2 elements
            self.snapshots = self.snapshots[-2:]

        old_snapshot = self.snapshots[-2]
        snapshot = self.snapshots[-1]
        kind = self.mode
        stats = snapshot.filter_traces(self.filters).compare_to(
                    old_snapshot.filter_traces(self.filters), kind)

        print('\n\n== Memory report ==')
        for stat in stats[:10]:
            fmt = "{} new {} total / {} new {} total memory blocks: "
            print(fmt.format(byteformat.format(stat.size_diff),
                             byteformat.format(stat.size),
                             stat.count_diff,
                             stat.count))
            for line in stat.traceback.format():
                print(line)
            print()

    def printTasksStats(self):
        tasks = asyncio.all_tasks()
        print('#{} tasks'.format(len(tasks)))

        if self.printAllTasks:
            for task in tasks:
                print(task)
                print()
        print()

    def analyzeGarbage(self):
        '''Garbage seems to be constant, so we should not worry about it

        # Add this at the start of your program
        import gc
        gc.set_debug(gc.DEBUG_SAVEALL)
        '''

        gc.collect()
        garbage = []
        for item in gc.garbage:
            garbage.append(sys.getsizeof(item))

        garbage.sort()
        print(garbage)
        print(sum(garbage))

    async def run(self):
        i = 0
        while True:
            # gc.collect()
            print('RSS: ' + byteformat.format(self.p.memory_info().rss))
            # Comment the next line to get a very simple mem print
            self.collect_stats()
            self.printTasksStats()

            if self.steps is not None and i == self.steps:
                return

            i += 1

            # Sleep
            await asyncio.sleep(self.duration)

            if self.stop:
                return

    def terminate(self):
        self.stop = True



'''Small class to record data attached to a connection

FIXME: use this to enforce publish+subscribe permissions.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import uuid


class ConnectionState:
    def __init__(self, appkey):
        self.appkey = appkey

        self.connection_id = uuid.uuid4().hex[:8]
        self.subscriptions = {}
        self.ok = True
        self.role = 'na'
        self.authenticated = False

        self.nonce = None
        self.error = 'na'

        self.path = f'/tmp/log_{self.connection_id}'
        self.fileLogging = False

    def log(self, msg):
        log = f"[{self.connection_id}::{self.role}] {msg}"
        print(log)

        if self.fileLogging:
            with open(self.path, 'a') as f:
                f.write(log + '\n')



'''Memoization decorator for functions taking one or more arguments.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''


def memoize(f):
    """Memoization decorator for functions taking one or more arguments."""

    class Memodict(dict):
        '''Helper class derived from dict'''

        def __init__(self, f):
            '''Constructor'''

            dict.__init__(self)
            self.f = f

        def __call__(self, *args):
            '''Call function'''

            return self[args]

        def __missing__(self, key):
            '''Missing method'''

            ret = self[key] = self.f(*key)
            return ret

    return Memodict(f)



'''Subscribe to a channel

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import click
import uvloop

from cobras.common.throttle import Throttle
from cobras.runner.superuser import preventRootUsage
from cobras.server.redis_connections import RedisConnections
from cobras.server.subscribe import runSubscriber


class MessageHandlerClass:
    def __init__(self, obj):
        self.cnt = 0
        self.cntPerSec = 0
        self.throttle = Throttle(seconds=1)

    def log(self, msg):
        print(msg)

    async def on_init(self, redisConnection):
        pass

    async def handleMsg(self, msg: str, payloadSize: int) -> bool:
        self.cnt += 1
        self.cntPerSec += 1

        if self.throttle.exceedRate():
            return True

        print(f"#messages {self.cnt} msg/s {self.cntPerSec}")
        self.cntPerSec = 0

        return True


@click.command()
@click.option('--redis_urls', default='redis://localhost')
@click.option('--redis_password')
@click.option('--channel')
@click.option('--appkey')
def redis_subscribe(redis_urls, redis_password, channel, appkey):
    '''Subscribe to a channel

    \b
    cobra subscribe --redis_urls redis://localhost:7379 --channel foo --appkey bar  # noqa

    \b
    cobra subscribe --redis_urls redis://redis:6379 --channel foo --appkey bar
    [From a bench container]

    \b
    cobra subscribe --redis_urls 'redis://localhost:7001;redis://localhost:7002' --channel foo --appkey bar  # noqa
    '''

    preventRootUsage()
    uvloop.install()

    appChannel = '{}::{}'.format(appkey, channel)
    redisConnections = RedisConnections(redis_urls, redis_password)
    runSubscriber(redisConnections, appChannel, MessageHandlerClass)



'''Client for a cobra server.

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import functools
import itertools
import json

import click
import websockets

from cobras.common.auth_hash import computeHash


class AuthException(Exception):
    pass


class HandshakeException(Exception):
    pass


async def clientInternal(url, creds, clientCallback):
    '''Main client. Does authenticate then invoke the clientCallback which
    takes control.
    '''

    async with websockets.connect(url) as websocket:

        idIterator = itertools.count()
        role = creds['role']

        handshake = {
            "action": "auth/handshake",
            "id": next(idIterator),
            "body": {
                "data": {
                    "role": role
                },
                "method": "role_secret"
            }
        }
        print(f"> {handshake}")
        await websocket.send(json.dumps(handshake))

        handshakeResponse = await websocket.recv()
        print(f"< {handshakeResponse}")

        response = json.loads(handshakeResponse)
        if response.get('action') != 'auth/handshake/ok':
            raise HandshakeException('Handshake error.')

        reply = json.loads(handshakeResponse)
        nonce = bytearray(reply['body']['data']['nonce'], 'utf8')

        secret = bytearray(creds['secret'], 'utf8')

        challenge = {
            "action": "auth/authenticate",
            "id": next(idIterator),
            "body": {
                "method": "role_secret",
                "credentials": {
                    "hash": computeHash(secret, nonce)
                }
            }
        }
        print(f"> {challenge}")
        await websocket.send(json.dumps(challenge))

        challengeResponse = await websocket.recv()
        print(f"< {challengeResponse}")

        response = json.loads(challengeResponse)
        if response.get('action') != 'auth/authenticate/ok':
            raise AuthException('Authentication error.')

        return await clientCallback(websocket)


async def client(url, creds, clientCallback):

    while True:
        try:
            return await clientInternal(url, creds, clientCallback)
        except TimeoutError as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except ConnectionRefusedError as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except ConnectionResetError as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except websockets.exceptions.ConnectionClosed as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except OSError as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass
        except AuthException as e:
            click.secho(str(e), fg='red')
            await asyncio.sleep(1)
            pass


async def subscribeHandler(websocket, **args):
    channel = args['channel']
    fsqlFilter = args['fsqlFilter']
    messageHandlerClass = args['messageHandlerClass']
    messageHandlerArgs = args['messageHandlerArgs']
    subscriptionId = args.get('subscription_id', channel)
    messageHandlerArgs['subscription_id'] = subscriptionId

    subscription = {
        "action": "rtm/subscribe",
        "body": {
            "subscription_id": subscriptionId,
            "channel": channel,
            "fast_forward": True,
            "filter": fsqlFilter
        },
        "id": 3  # FIXME
    }
    print(f"> {subscription}")
    await websocket.send(json.dumps(subscription))

    subscribeResponse = await websocket.recv()
    print(f"< {subscribeResponse}")

    messageHandler = messageHandlerClass(websocket, messageHandlerArgs)
    await messageHandler.on_init()

    async for msg in websocket:
        ret = await messageHandler.handleMsg(msg)
        if not ret:
            break

    return messageHandler


async def subscribeClient(url, credentials, channel,
                          fsqlFilter, messageHandlerClass,
                          messageHandlerArgs):
    subscribeHandlerPartial = functools.partial(
        subscribeHandler, channel=channel,
        fsqlFilter=fsqlFilter, messageHandlerClass=messageHandlerClass,
        messageHandlerArgs=messageHandlerArgs)

    ret = await client(url, credentials, subscribeHandlerPartial)
    return ret



'''Module used for python packaging

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''
from __future__ import absolute_import

import platform
import os
import sys
from setuptools import find_packages, setup

if sys.version_info[:2] < (3, 6):
    print('Error: Cobra requires Python 3.6')
    sys.exit(1)

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))

with open(os.path.join(ROOT, 'DOCKER_VERSION')) as f:
    VERSION = f.read().strip()


dev_requires = [
    'flake8',
    'isort',
    'honcho',
    'mypy',
    'aiojobs==0.1.0',
    'twine'
]

tests_require = [
    'pytest',
    'pytest-cov',
    'pytest-xdist',
    'coverage'
]

with open('requirements.txt') as f:
    install_requires = f.read().splitlines()

setup(
    name='cobras',
    version=VERSION,
    author='Benjamin Sergeant',
    author_email='bsergean@gmail.com',
    url='https://github.com/machinezone/cobra',
    description='A realtime messaging server using WebSockets and Redis.',
    long_description=open(os.path.join(ROOT, 'README.md')).read(),
    long_description_content_type='text/markdown',
    package_dir={'': 'src'},
    packages=find_packages('src'),
    zip_safe=False,
    install_requires=install_requires,
    extras_require={
        'tests': tests_require,
        'dev': dev_requires
    },
    license='BSD 3',
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'cobra = cobras.runner:cli'
        ],
    },
    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Operating System :: POSIX :: Linux',
        'Topic :: Software Development'
    ],
)



'''Misc algorithm routines

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''


def transpose(entries):
    '''Transpose a matrix'''

    # Lame fix
    if isinstance(entries, list) and len(entries) == 1 and entries[0] is None:
        return []

    return list(map(list, zip(*entries)))



'''Subscribe to custom channels that let us monitor cobra

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio
import collections
import json

import byteformat
import click
import tabulate

from cobras.client.client import subscribeClient
from cobras.common.algorithm import transpose
from cobras.common.throttle import Throttle
from cobras.server.stats import DEFAULT_STATS_CHANNEL


def writeJson(data):
    '''JSON Pretty printer'''

    return json.dumps(data, sort_keys=True,
                      indent=4, separators=(',', ': '))


class MessageHandlerClass:
    def __init__(self, websockets, args):
        self.cnt = 0
        self.throttle = Throttle(seconds=1)
        self.resetMetrics()

        self.raw = args['raw']
        self.roleFilter = args['role_filter']
        self.showNodes = args['show_nodes']
        self.subscribers = args['subscribers']

        self.roleMetrics = []

    def resetMetrics(self):
        self.metrics = collections.defaultdict(int)
        self.nodes = set()

        self.nodeEntries = []
        self.nodeEntriesHeader = None

        self.allRoleMetrics = {}
        self.roles = set()

    async def on_init(self):
        pass

    def humanReadableSize(self, key, val):
        return byteformat.format(val) if '_bytes' in key else val

    def shouldProcessNode(self, node):
        return 'subscriber' in node if self.subscribers else True

    async def handleMsg(self, msg: str) -> bool:

        data = json.loads(msg)['body']['messages'][0]
        node = data['node']

        if node not in self.nodes and self.shouldProcessNode(node):
            # nodeEntry = [data['node'][:8]]
            nodeEntry = [node]
            nodeEntryHeaders = ['Nodes']

            cobraData = data['data']['cobra']
            for key in sorted(cobraData.keys()):
                s = sum(cobraData[key].values())
                self.metrics[key] += s

                s = self.humanReadableSize(key, s)

                nodeEntry.append(s)
                nodeEntryHeaders.append(key)

            self.updateRoleMetrics(cobraData)

            # System stats
            for metric in data['data']['system'].keys():
                val = data['data']['system'][metric]

                if metric == 'connections':
                    self.metrics[metric] += val

                nodeEntryHeaders.append(metric)

                val = self.humanReadableSize(metric, val)

                nodeEntry.append(val)

            self.nodeEntries.append(nodeEntry)

            self.nodes.add(node)

            if self.nodeEntriesHeader is None:
                self.nodeEntriesHeader = nodeEntryHeaders

        if self.throttle.exceedRate():
            return True

        click.clear()
        # print(yaml.dump(data))

        self.metrics = {key: self.humanReadableSize(key, val)
                        for key, val in self.metrics.items()}

        print(writeJson(self.metrics))
        print()

        # Print a table with all nodes
        nodeEntries = [self.nodeEntriesHeader]
        self.nodeEntries.sort()
        nodeEntries.extend(self.nodeEntries)

        # Transpose our array and print the nodes horizontally
        if 0 < len(nodeEntries) < 8:
            nodeEntries = transpose(nodeEntries)

        if self.showNodes:
            print(tabulate.tabulate(nodeEntries,
                                    tablefmt="simple",
                                    headers="firstrow"))

        self.displayRoleMetrics()

        self.resetMetrics()
        return True

    def updateRoleMetrics(self, cobraData):
        '''Collect data per role'''

        for metric, metricData in cobraData.items():

            metricByRole = self.allRoleMetrics.get(metric)
            if metricByRole is None:
                metricByRole = collections.defaultdict(int)

            for role, val in metricData.items():
                metricByRole[role] += val

                if self.roleFilter is not None:
                    if self.roleFilter in role:
                        self.roles.add(role)
                else:
                    self.roles.add(role)

            self.allRoleMetrics[metric] = metricByRole

        if self.raw:
            print(cobraData)

    def displayRoleMetrics(self):
        '''Display data broken down per role'''

        rows = [['Roles'] + list(sorted(self.roles))]

        for metric in sorted(self.allRoleMetrics):
            metricByRole = self.allRoleMetrics[metric]
            # print(metric, metricByRole)

            row = [metric]

            for role in sorted(self.roles):
                val = metricByRole.get(role, 0)

                val = self.humanReadableSize(metric, val)

                row.append(val)

            rows.append(row)

        print()
        print(tabulate.tabulate(rows,
                                tablefmt="simple",
                                headers="firstrow"))


def runMonitor(url, credentials, raw, roleFilter, showNodes, subscribers):
    asyncio.get_event_loop().run_until_complete(
            subscribeClient(url, credentials, DEFAULT_STATS_CHANNEL,
                            '', MessageHandlerClass,
                            {'raw': raw,
                             'role_filter': roleFilter,
                             'show_nodes': showNodes,
                             'subscribers': subscribers}))



'''Utility to make sure that exceptions are handled and logged to sentry

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import asyncio

from sentry_sdk import capture_exception


def addTaskCleanup(task):
    def exception_logging_done_cb(fut):
        try:
            e = fut.exception()
        except asyncio.CancelledError:
            return

        if e is not None:
            # Sentry
            capture_exception(e)

            # Normal exception business
            asyncio.get_event_loop().call_exception_handler({
                'message': 'Unhandled exception in async future',
                'future': fut,
                'exception': e,
            })

    task.add_done_callback(exception_logging_done_cb)



'''Run the cobra frontend

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''

import os

import click
import sentry_sdk
import uvloop

from cobras.common.apps_config import getDefaultAppsConfigPath
from cobras.runner.superuser import preventRootUsage
from cobras.server.app import AppRunner


@click.command()
@click.option('--host', envvar='COBRA_HOST', default='127.0.0.1',
              help='Binding host address. Set to 0.0.0.0 in prod environments')
@click.option('--port', envvar='COBRA_PORT', default='8765')
@click.option('--redis_urls', envvar='COBRA_REDIS_URLS',
              default='redis://localhost;redis://localhost')
@click.option('--redis_password', envvar='COBRA_REDIS_PASSWORD')
@click.option('--apps_config_path', envvar='COBRA_APPS_CONFIG',
              default=getDefaultAppsConfigPath())
@click.option('--verbose', envvar='COBRA_VERBOSE', is_flag=True)
@click.option('--prod', envvar='COBRA_PROD', is_flag=True)
@click.option('--plugins', envvar='COBRA_PLUGINS')
@click.option('--debug_memory', envvar='COBRA_DEBUG_MEMORY', is_flag=True)
@click.option('--sentry', envvar='COBRA_SENTRY', is_flag=True)
@click.option('--sentry_url', envvar='COBRA_SENTRY_URL')
@click.option('--no_stats', envvar='COBRA_NO_STATS', is_flag=True)
@click.option('--max_subscriptions', envvar='COBRA_MAX_SUSBSCRIPTIONS',
              default=-1)
@click.option('--idle_timeout', envvar='COBRA_IDLE_TIMEOUT',
              default=5 * 60,
              help='idle connections kicked out after X seconds')
def run(host, port, redis_urls, redis_password,
        apps_config_path, verbose, debug_memory, plugins,
        sentry, sentry_url, prod, no_stats, max_subscriptions, idle_timeout):
    '''Run the cobra server

    \b
    cobra run --redis_urls 'redis://localhost:7001;redis://localhost:7002'
    \b
    env COBRA_REDIS_PASSWORD=foobared cobra run
    '''
    if prod:
        os.environ['COBRA_PROD'] = '1'

    preventRootUsage()
    uvloop.install()

    if sentry and sentry_url:
        sentry_sdk.init(sentry_url)

    print('runServer', locals())
    runner = AppRunner(host, port, redis_urls,
                       redis_password, apps_config_path,
                       verbose, debug_memory, plugins,
                       not no_stats, max_subscriptions, idle_timeout)
    runner.run()



'''Main entry point for the cobra websocket server

Copyright (c) 2018-2019 Machine Zone, Inc. All rights reserved.
'''
import asyncio
import datetime
import functools
import http
import importlib
import logging
import time
import traceback

import websockets

from cobras.common.apps_config import STATS_APPKEY, AppsConfig
from cobras.common.memory_debugger import MemoryDebugger
from cobras.common.task_cleanup import addTaskCleanup
from cobras.common.version import getVersion
from cobras.server.connection_state import ConnectionState
from cobras.server.pipelined_publishers import PipelinedPublishers
from cobras.server.protocol import parseAppKey, processCobraMessage
from cobras.server.redis_connections import RedisConnections
from cobras.server.stats import ServerStats


async def cobraHandler(websocket,
                       path,
                       app,
                       redisUrls: str,
                       verbose: bool):
    start = time.time()
    msgCount = 0
    appkey = parseAppKey(path)  # appkey must have been validated

    state: ConnectionState = ConnectionState(appkey)
    state.log('appkey {}'.format(state.appkey))

    key = state.connection_id
    app['connections'][key] = (state, websocket)

    app['stats'].incrConnections(appkey)
    connectionCount = len(app['connections'])
    state.log(f'(open) connections {connectionCount}')

    try:
        async for message in websocket:
            msgCount += 1
            await processCobraMessage(state, websocket, app, message)
            if not state.ok:
                raise Exception(state.error)

    except websockets.exceptions.WebSocketProtocolError as e:
        print(e)
        state.log('Protocol error')
    except websockets.exceptions.ConnectionClosed:
        state.log('Connection closed')
    except Exception as e:
        print(e)
        print('Generic Exception caught in {}'.format(traceback.format_exc()))
    finally:
        del app['connections'][state.connection_id]

        subCount = len(state.subscriptions)

        if subCount > 0:
            state.log('cancelling #{} subscriptions'.format(subCount))
        for val in state.subscriptions.values():
            task, role = val
            app['stats'].decrSubscriptionsBy(role, 1)
            task.cancel()

        uptime = time.time() - start
        uptimeStr = str(datetime.timedelta(seconds=uptime))
        uptimeStr, _, _ = uptimeStr.partition('.')  # skip the milliseconds

        status = f'(close) uptime {uptimeStr} msgcount {msgCount}'
        status += ' connections {}'.format(len(app['connections']))
        state.log(status)

        app['stats'].decrConnections(appkey)


class ServerProtocol(websockets.WebSocketServerProtocol):
    '''Used to validate appkey'''
    appsConfig = None

    async def process_request(self, path, request_headers):

        if path == '/health/':
            return http.HTTPStatus.OK, [], b'OK\n'

        if path == '/version/':
            return http.HTTPStatus.OK, [], bytes(getVersion(), 'utf8') + b'\n'

        appkey = parseAppKey(path)
        if appkey is None or not ServerProtocol.appsConfig.isAppKeyValid(appkey):  # noqa
            return http.HTTPStatus.FORBIDDEN, [], b'KO\n'


class AppRunner():
    '''From aiohttp
    '''

    def __init__(self, host, port, redisUrls,
                 redisPassword, appsConfigPath, verbose,
                 debugMemory, plugins, enableStats,
                 maxSubscriptions, idleTimeout):
        self.app = {}
        self.app['connections'] = {}
        self.app['apps_config_path'] = appsConfigPath
        self.app['max_subscriptions'] = maxSubscriptions
        self.app['idle_timeout'] = idleTimeout

        self.app['verbose'] = verbose
        self.app['memory_debugger'] = debugMemory
        self.app['redis_urls'] = redisUrls
        self.app['redis_password'] = redisPassword

        self.host = host
        self.port = port
        self.redisUrls = redisUrls
        self.redisPassword = redisPassword
        self.verbose = verbose
        self.plugins = plugins
        self.enableStats = enableStats

        appsConfig = AppsConfig(appsConfigPath)
        self.app['apps_config'] = appsConfig

        try:
            appsConfig.validateConfig()
        except ValueError as e:
            logging.error(f'Invalid apps config file: {e}')
            pass

        try:
            if plugins is not None:
                self.app['plugins'] = importlib.import_module(plugins)
        except ImportError:
            logging.error(f'failure to import {plugins}')

        self.app['batch_publish_size'] = appsConfig.getBatchPublishSize()
        self.server = None

    async def init_app(self):
        '''Example urls:
           * redis://localhost
           * redis://redis
           * redis://172.18.176.220:7379
           * redis://sentryredis-1-002.shared.live.las1.mz-inc.com:6310
        '''
        redisUrls = self.app['redis_urls']
        redisPassword = self.app['redis_password']
        batchPublishSize = self.app['batch_publish_size']
        redisConnections = RedisConnections(redisUrls, redisPassword)

        pipelinedPublishers = PipelinedPublishers(redisConnections,
                                                  batchPublishSize)
        self.app['pipelined_publishers'] = pipelinedPublishers

        serverStats = ServerStats(pipelinedPublishers, STATS_APPKEY)
        self.app['stats'] = serverStats

        if self.enableStats:
            self.serverStatsTask = asyncio.create_task(serverStats.run())
            addTaskCleanup(self.serverStatsTask)

        if self.app.get('memory_debugger'):
            memoryDebugger = MemoryDebugger()
            self.app['memory_debugger'] = memoryDebugger

            self.memoryDebuggerTask = asyncio.create_task(memoryDebugger.run())
            addTaskCleanup(self.memoryDebuggerTask)

    async def cleanup(self):
        # FIXME: we could speed this up
        self.app['stats'].terminate()
        await self.serverStatsTask

        if self.app.get('memory_debugger'):
            self.app['memory_debugger'].terminate()
            await self.memoryDebuggerTask

    async def setup(self):
        await self.init_app()

        handler = functools.partial(cobraHandler,
                                    app=self.app,
                                    redisUrls=self.redisUrls,
                                    verbose=self.verbose)

        ServerProtocol.appsConfig = self.app['apps_config']

        self.server = await websockets.serve(handler,
                                             self.host,
                                             self.port,
                                             create_protocol=ServerProtocol,
                                             subprotocols=['json'],
                                             ping_timeout=None,
                                             ping_interval=None)

    def run(self):
        asyncio.get_event_loop().run_until_complete(self.setup())
        asyncio.get_event_loop().run_forever()

        # Is that ever reached ?
        # When I close the server with Ctrl-C click
        # might terminate the process right away.
        # Regardless self.terminate is needed for unittesting.
        # We could call it in client code only ?
        self.terminate()

    def closeRedis(self):
        # FIXME (is that sufficient to close redis ?)
        db = self.app['redis']
        db.close()
        asyncio.ensure_future(db.wait_closed())

    def terminate(self):
        asyncio.get_event_loop().run_until_complete(self.cleanup())

        # Now close websocket server
        self.server.close()
        self.server.wait_closed()
